#! /usr/bin/env python
# -*- coding: utf-8 -*-
####################
# NuVo Grand Concerto Plugin for Indigo
# Developed by Jeremy Swancoat

import os
import sys
import time
import serial
import requests
import socket

# Note the "indigo" module is automatically imported and made available inside
# our global name space by the host process. 

##################################################################################################
class Plugin(indigo.PluginBase):

	##############################################################################################
	# Required Methods
	##############################################################################################
	##############################################################################################
	def __init__(self, pluginId, pluginDisplayName, pluginVersion, pluginPrefs):
		# Call the base class's init method
		indigo.PluginBase.__init__(self, pluginId, pluginDisplayName, pluginVersion, pluginPrefs)
		# Setting debug to True will show you verbose debugging information in the Indigo Event Log
		if 'showDebugInfo' in pluginPrefs:
			self.debug = pluginPrefs['showDebugInfo']
		else:
			self.debug = False
		self.conn = None
		self.portEnabled = False
		self.commQueue = []
		# self.zonedev is a dictionary of the Zone Identifier and the device using it.
		self.zonedev = {}
		# self.inputdev is a dictionary of the Source Identifier and the device using it.
		self.inputdev = {}
		self.inputnuvonet = {}
		self.inputname = {}
		self.inputlastst = {}
		self.inputremote = {}
		# self.spDir is a dictionary of Shairport Device IDs and the Source they are assigned to.
		self.spDir = {}
		self.varDir = {}
		self.kShairportId = "com.jeremyswancoat.indigoplugin.shairportsync"
		
		#self.inputname needs it's keys initialized at least at the start...
		self.inputname['S1'] = ""
		self.inputname['S2'] = ""
		self.inputname['S3'] = ""
		self.inputname['S4'] = ""
		self.inputname['S5'] = ""
		self.inputname['S6'] = ""
		
		indigo.variables.subscribeToChanges()
					
	########################################
	def __del__(self):
		indigo.PluginBase.__del__(self)
		
	########################################
	def startup(self):
		# startup process for the serial device. Other than open the serial port, there should be none really. It just runs.
		self.logger.debug(u"Startup Called")
		
		if self.conn is None:
			pass
		else:
			self.conn.close()		
		self.portEnabled = False
		

		serialUrl = self.getSerialPortUrl(self.pluginPrefs, u"serialport")
		if serialUrl == "":
			pass
		else:
			self.logger.info("Serial Port URL is " + serialUrl)
			self.conn = self.openSerial(u"NuVo Grand Concerto", serialUrl, 57600, stopbits=1, timeout=0.1, writeTimeout=1)
			
		if self.conn is not None:
			self.portEnabled = True
			self.logger.info("Serial Port Open at " + serialUrl)
		
		indigo.server.subscribeToBroadcast(self.kShairportId,u"startingBcast",u"startingBcast")
		indigo.server.subscribeToBroadcast(self.kShairportId,u"endingBcast",u"endingBcast")
		indigo.server.subscribeToBroadcast(self.kShairportId,u"ShairportData",u"receiveBcast")
		
		for s in range(1, 7):
			srclbl = "S" + str(s)
			self.commQueue.append(srclbl + "ACTIVE?")
			self.commQueue.append(srclbl + "NAME?")
		
	def startingBcast(self):
		self.logger.info("Shairport Broadcast Starting")
	
	def endingBcast(self):
		self.logger.info("Shairport Broadcast Stopping")
		
	def receiveBcast(self, arg):
		self.logger.debug("-----------Broadcast Received----------")
		self.logger.debug("Dev ID: " + str(arg[0]))
		self.logger.debug("State: " + str(arg[1]))
		if arg[0] in self.spDir:
			self.logger.debug(str(arg[0]) + " found in spDir")
			inputcode = self.spDir[arg[0]]
			self.logger.debug(str(arg[0]) + " assigned to " + self.inputdev[inputcode].name)
			if inputcode in self.inputdev:
				stateid = arg[1]
				mdata = arg[2]
				if stateid == 'nowplaying':
					if mdata == True:
						self.inputlastst[inputcode] = 0
					else:
						self.inputlastst[inputcode] = round(time.time(),0)
						servdev = self.inputdev[inputcode]
						if servdev.pluginProps["repTrackTime"] == True:
							length = 10 * self.timeToSec(servdev.states["tracklen"])
							prog = 10 * self.timeToSec(servdev.states["trackprog"])
						else:
							length = 0
							prog = 0
						self.commQueue.append(inputcode + "DISPINFO," + str(length) + "," + str(prog) + ",0")
				elif stateid == 'timeDict':
					if mdata['nowplaying'] == True:
						zstate = 2
					else:
						zstate = 0
					length = 10 * int(mdata['tracklen'])
					prog = 10 * int(mdata['progsec'])
					self.commQueue.append(inputcode + "DISPINFO," + str(length) + "," + str(prog) + "," + str(zstate))
				elif stateid == 'clearData':
					self.resetSourceMetaData(inputcode)
				elif stateid == 'bcastDict':
					dev = self.inputdev[inputcode]
					for Ln in range(1, 5):
						primst = dev.pluginProps["spLine" + str(Ln) + "p"]
						secst = dev.pluginProps["spLine" + str(Ln) + "s"]
						primd = mdata[primst]
						secd = mdata[secst]
						if primst == "none":
							scdata = ""
						elif primd != "":
							scdata = primd
						else:
							scdata = secd
						self.commQueue.append(inputcode + "DISPLINE" + str(Ln) + "\"" + scdata + "\"")
					self.inputremote[inputcode]['clientip'] = mdata['clientip']
					self.inputremote[inputcode]['srcremport'] = mdata['srcremport']
					self.inputremote[inputcode]['artoken'] = mdata['artoken']
					self.inputremote[inputcode]['dacpid'] = mdata['dacpid']
			
					
	def variableUpdated(self, origVar, newVar):
		varid = str(newVar.id)
		if varid in self.varDir:
			inputcode = self.varDir[varid]
			dev = self.inputdev[inputcode]
			newvalue = str(newVar.value)
			for Ln in range(1, 4):
				if dev.pluginProps["vLine" + str(Ln)] == varid:
					self.commQueue.append(inputcode + "DISPLINE" + str(Ln) + "\"" + newvalue + "\"")
		
	
	def shutdown(self):
		# close serial port here
		self.logger.debug(u"Shutdown Called")
		if self.conn is None:
			pass
		else:
			self.conn.close()
		self.logger.info("Serial Port Closed")
		
	def genSpDevs(self, filter, valuesDict, typeID, targetID):
		devlist = []
		for dev in indigo.devices.iter(self.kShairportId + ".shairport"):
			devlist.append((dev.id, dev.name))
		return devlist

	def genSpMdList(self, filter, valuesDict, typeID, targetID):
		sourcelist = []
		sourcelist.append(("none","None (Leave Blank)"))
		sourcelist.append(("added","Date Added"))
		sourcelist.append(("album","Album"))
		sourcelist.append(("artist","Artist"))
		sourcelist.append(("albumartist","Album Artist"))
		sourcelist.append(("bitrate","Bitrate"))
		sourcelist.append(("bpm","BPM"))
		sourcelist.append(("comment","Comment"))
		sourcelist.append(("devicename","Device Name"))
		sourcelist.append(("filedesc","File Description"))
		sourcelist.append(("format","Format"))
		sourcelist.append(("genre","Genre"))
		sourcelist.append(("modified","Date Modified"))
		sourcelist.append(("name","Title"))
		sourcelist.append(("podcast","Podcast (T/F)"))
		sourcelist.append(("samplerate","Sample Rate"))
		sourcelist.append(("trackcount","Track Count"))
		sourcelist.append(("tracknumber","Track Number"))
		sourcelist.append(("useragent","User Agent"))
		sourcelist.append(("year","Year"))				
		return sourcelist
		
	def genVarMdList(self, filter, valuesDict, typeID, targetID):
		sourcelist = []
		sourcelist.append(("none","None (Leave Blank)"))
		for var in indigo.variables:
			sourcelist.append((var.id, var.name))
		return sourcelist
	
	def validateDeviceConfigUi(self, valuesDict, typeId, devId):
		# Validate that appropriate selections were made in device creation / modification dialogs.
		# Ensure that Zone is not already used
		errorsDict = indigo.Dict()
		self.logger.debug("Validating Device")
		self.logger.debug("Type ID: " + typeId)
		
		if typeId == "zone":
			self.logger.debug("Zone Selected: " + valuesDict["zoneselect"])
			if valuesDict["zoneselect"] in self.zonedev:
				if self.zonedev[valuesDict["zoneselect"]].id == devId:
					self.logger.debug("Zone Available")
				else:
					errorsDict["zoneselect"] = "Zone Already Assigned to Device"
					
		if typeId == "source":
			self.logger.debug("Source Selected: " + valuesDict["inputselect"])
			if valuesDict["inputselect"] in self.inputdev:
				if self.inputdev[valuesDict["inputselect"]].id == devId:
					self.logger.debug("Input Available")
				else:
					errorsDict["inputselect"] = "Input Already Assigned to Device"
				if valuesDict["supplyMd"] == True:
					if self.inputnuvonet[valuesDict["inputselect"]] == True:
						errorsDict["supplyMd"] = "Source is set up as a NuVo Net source. Cannot supply metadata to this type of source."
					if valuesDict["metadataSourceType"] == "shairport":
						spdeviceid = valuesDict["spDevice"]
						if spdeviceid in self.spDir:
							if self.spDir[spdeviceid] != valuesDict["inputselect"]:
								errorsDict["spDevice"] = "Shairport Device already in use by another source"
						try:
							clearsecint = int(valuesDict["clearsec"])
							if clearsecint < 0:
								errorsDict["clearsec"] = "Time to Clear Data must be a number greater than zero"
						except:
							errorsDict["clearsec"] = "Time to Clear Data must be a number greater than zero"
													
		if len(errorsDict) > 0:
			return (False, valuesDict, errorsDict)
		else:
			return (True, valuesDict)

	def deviceStartComm(self, dev):
		self.logger.debug("deviceStartComm called")
		if dev.model == "Input Source":
			if dev.pluginProps["supplyMd"] == True:
				if dev.pluginProps["metadataSourceType"] == "shairport":
					self.spDir[dev.pluginProps["spDevice"]] = dev.pluginProps["inputselect"]
					self.inputremote[dev.pluginProps["inputselect"]] = {}
				elif dev.pluginProps["metadataSourceType"] == "variables":
					for v in range(1, 5):
						vLine = "vLine" + str(v)
						vid = dev.pluginProps[vLine]
						if vid != "none":
							self.varDir[vid] = dev.pluginProps["inputselect"]
			self.inputdev[dev.pluginProps["inputselect"]] = dev
			self.inputlastst[dev.pluginProps["inputselect"]] = 0
			self.logger.debug("Just added: " + dev.pluginProps["inputselect"] + " to inputdev")
			self.logger.debug("Calling Status: " + dev.pluginProps["inputselect"] + "DISPINFO?")
			self.commQueue.append(dev.pluginProps["inputselect"] + "DISPINFO?")
			self.logger.debug("Calling Status: " + dev.pluginProps["inputselect"] + "DISPLINE?")
			self.commQueue.append(dev.pluginProps["inputselect"] + "DISPLINE?")
		else:
			self.zonedev[dev.pluginProps["zoneselect"]] = dev
			self.logger.debug("Just added: " + dev.pluginProps["zoneselect"] + " to zonedev")
			self.logger.debug("Calling Status: " + dev.pluginProps["zoneselect"] + "STATUS?")
			self.commQueue.append(dev.pluginProps["zoneselect"] + "STATUS?")
		
	def deviceStopComm(self, dev):
		if dev.model == "Input Source":
			del self.inputdev[dev.pluginProps["inputselect"]]
			del self.inputlastst[dev.pluginProps["inputselect"]]
			if dev.pluginProps["supplyMd"] == True:
				if dev.pluginProps["metadataSourceType"] == "shairport":
					del self.spDir[dev.pluginProps["spDevice"]]
					del self.inputremote[dev.pluginProps["inputselect"]]
				elif dev.pluginProps["metadataSourceType"] == "variables":
					for v in range(1, 4):
						vLine = "vLine" + str(v)
						vid = dev.pluginProps[vLine]
						if vid in self.varDir:
							del self.varDir[vid]
			self.logger.debug("Just deleted: " + dev.pluginProps["inputselect"] + " from inputdev")
		else:
			del self.zonedev[dev.pluginProps["zoneselect"]]
			self.logger.debug("Just deleted: " + dev.pluginProps["zoneselect"] + " from zonedev")
		
	def validatePrefsConfigUi(self, valuesDict):
		
		errorsDict = indigo.Dict()		
		self.logger.debug("Validating Serial Port")
		self.validateSerialPortUi(valuesDict, errorsDict, u"serialport")
		
		if valuesDict[u'showDebugInfo'] == True:
			self.debug = True
		else:
			self.debug = False				
		
		if len(errorsDict) > 0:
			return (False, valuesDict, errorsDict)
		return (True, valuesDict)

		
	def closedPrefsConfigUi(self, valuesDict, userCancelled):
		# Basically, when the Config box closes, we want another chance to open the serial port.
		# Since that's all the startup method does, just call it from here.
		self.logger.debug("closedPrefsConfigUI() called")
		
		#Obviously, if the box closes by cancelling, then do nothing.
		if userCancelled:
			pass
		else:
			self.startup()
		
	def runConcurrentThread(self):
		## This is what runs as the plugin is simply 'running'. 
		## Basically, if a command comes in from the serial device, you read it, parse it and hand it off to
		## whatever device/or variable needs to have its state updated.
		## In the meantime, you have a queue set up which gets loaded with commands from devices which must be fed
		## to the serial interface to be enacted.
		sleeptime = 0.1
		lastupdate = round(time.time(),0)
		try:
			while True:
				if self.portEnabled:
					#self.logger.debug("Loop Running")
					sleeptime = 0.1
					fromgrandconcerto = 0
					fromgrandconcerto = self.conn.readline()
					if len(fromgrandconcerto) > 3:
						sleeptime = 0.001
						self.logger.debug("Nuvo: " + fromgrandconcerto[:-2])  #removes last two characters (CR and LF)
						self.parseToServer(fromgrandconcerto[:-2])
					if len(self.commQueue) > 0:
						sleeptime = 0.001
						# Sleep is set slow to keep processor use down, but putting 18 commands through with
						# 0.1 sleeptime takes almost 2 seconds, so if there's commands in the queue, drop the sleeptime, and
						# bang them through really fast. Then reset at beginning of loop.
						self.logger.debug(self.commQueue[0])
						sendcount = self.conn.write("*" + self.commQueue.pop(0) + "\r")	
					curstamp = round(time.time(),0)
					if curstamp > lastupdate:
						self.refreshTrackTimes()
						lastupdate = round(time.time(),0)
					for inputcode in self.inputlastst:
						if self.inputdev[inputcode].states['mData3'] == self.inputname[inputcode]:
							pass
						else:
							lasttime = self.inputlastst[inputcode]
							clearsec = int(self.inputdev[inputcode].pluginProps["clearsec"])
							if lasttime == 0:
								pass
							elif curstamp-lasttime > clearsec:
								self.resetSourceMetaData(inputcode)
								self.inputlastst[inputcode] = 0
								sleeptime = 0.001
							
				self.sleep(sleeptime)
		except self.StopThread:
			pass	
		
	def stopConcurrentThread(self):
		self.stopThread = True
		
	def menuGetSources(self, filter, valuesDict, typeID, devID):
		sourcelist = []
		for inputcode, inputname in self.inputname.iteritems():
			if inputcode in self.inputdev:
				inputname = self.inputdev[inputcode].name
			longinputcode = inputcode.replace("S","SRC")
			sourcelist.append((longinputcode,inputname))
			
		return sourcelist
		
	def menuGetZones(self, filter, valuesDict, typeID, devID):
		zonelist = []
		for zonecode, servdev in self.zonedev.iteritems():
			zonelist.append((zonecode,servdev.name))
		return zonelist
		
	def resetSourceMetaData(self, inputcode):
		self.logger.debug("Resetting Metadata on " + self.inputdev[inputcode].name)
		self.commQueue.append(inputcode + "DISPLINE1\"\"")
		self.commQueue.append(inputcode + "DISPLINE2\"\"")
		self.commQueue.append(inputcode + "DISPLINE3\"" + self.inputname[inputcode] + "\"")
		self.commQueue.append(inputcode + "DISPLINE4\"\"")
		self.commQueue.append(inputcode + "DISPINFO,0,0,0")
		
	def parseToServer(self, fromgc):
		#Somewhat Messy, but just has to test to see what kind of message it is, and then parse correctly.
		#All responses start with a '#', so start by stripping that.
		fromgc = fromgc[1:]
		#For purposes here, server responses fall into 4 main categories:
		#Zone statements start with a Z
		#Source statements start with an S
		#All Zones Off is simply ALLOFF
		#and then there are miscellaneous system commands we aren't interested in for now.
		if fromgc == "ALLOFF":
			for zonecode, servdev in self.zonedev.iteritems():
				self.logger.debug("All Off - " + zonecode + " called")
				servdev.updateStateOnServer("onOffState", value="off")
		elif fromgc[0] == "Z":
			#Zone statments start with Z.
			zonecode = fromgc[:2]
			if zonecode in self.zonedev:
				#If we're using this device we'll parse and process. If not, log and move on.
				self.logger.debug(zonecode + " found in devices list")
				servdev = self.zonedev[zonecode]
				if "," in fromgc:
					#on, off, source, volume type statements come in a zone delimited list. Zone actions, like button presses do not.
					resplist = fromgc.split(",")
					if resplist[1] == "OFF":
						#Off statments are simple and only have two elements in the list. Zone and off.
						servdev.updateStateOnServer("onOffState", value=False)
						self.logger.info(servdev.name + " is Off")
					elif resplist[1] == "ON":
						#Statements for active zones have a bunch of info. Pick it out and send to server Easy!
						#But first, save the current zone parameters, so you can compare and only put changes into the Event log.
						#This prevents the log from filling full of on/off states and source changes every time you change the volume.
						savedstate = servdev.states["onOffState"]
						savedinput = servdev.states["Input"]
						savedvolume = servdev.states["Volume"]
						#servdev.updateStateOnServer("onOffState", value="on")
						#servdev.updateStateOnServer("Input", value = resplist[2])
						inputcode = resplist[2].replace("SRC","S")
						if inputcode in self.inputdev:
							sname = self.inputdev[inputcode].name
						else:
							sname = self.inputname[inputcode]
							
						if resplist[3] == "MUTE":
							pass
						else:
							calcvol = 79 - int(resplist[3][-2:])
							servdev.updateStateOnServer("Volume", value=calcvol)
						if savedstate == False:
							self.logger.debug("savedstate = off")
							#if the zone is turning ON and WAS OFF, then report everything...
							#This is probably the only time you report the zone status as 'ON'
							servdev.updateStateOnServer("onOffState", value=True)
							self.logger.info(servdev.name + " is ON")
							servdev.updateStateOnServer("Input", value = resplist[2])
							self.logger.info(servdev.name + " source is " + sname)
							#self.logger.info(servdev.name + " source is " + self.pluginPrefs[resplist[2]])
							if resplist[3] == "MUTE":
								self.logger.info(servdev.name + " is muted")
							else:
								self.logger.debug("Reported Volume is " + resplist[3])
								self.logger.info(servdev.name + " volume is " + str(calcvol))
						else:
							#Now if the zone was already on, test the parameters, if it's unchanged, don't log
							if savedinput == resplist[2]:
								pass
							else:
								servdev.updateStateOnServer("Input", value = resplist[2])
								self.logger.info(servdev.name + " source is " + sname)
								#self.logger.info(servdev.name + " source is " + self.pluginPrefs[resplist[2]])
							if savedvolume == calcvol:
								pass
							else:
								if resplist[3] == "MUTE":
									self.logger.info(servdev.name + " volume is muted")
								else:
									self.logger.info(servdev.name + " volume is " + str(calcvol))
						savedstate = None
						savedinput = None
						savedvolume = None						
				else:
					inputcode = fromgc[2:4]
					command = fromgc[4:]
					if command[:4] == "MACRO":
						self.debug.info("Macro " + command[5:] + " called from " + servdev.name + ". (Macros not supported by Plugin)")
					else:
						self.commandFromZone(zonecode, inputcode, command)
						#self.logger.debug("zonecode: " + zonecode)
						#self.logger.debug("inputcode: " + inputcode)
						#self.logger.debug("command: " + command)
			else:
				self.logger.debug(zonecode + " is not yet assigned to a device.")
		elif fromgc[0] == "S":
			#Source statements start with S.
			inputcode = fromgc[:2]
			#Some Source statements we care about if the source is in use or not (device validation, etc.)
			if fromgc[2:8] == "ACTIVE":
				if int(fromgc[-1:]) == 1:
					self.inputnuvonet[inputcode] = True
					self.logger.debug(inputcode + " is Nuvo Net Source")
				else:
					self.inputnuvonet[inputcode] = False
					self.logger.debug(inputcode + " is not Nuvo Net Source")
			if fromgc[2:6] == "NAME":
				sysname = fromgc[6:].replace("\"","")
				self.inputname[inputcode] = sysname
				self.logger.debug(inputcode + " system Name is: " + sysname)
			# Statements relevant to active sources.
			if inputcode in self.inputdev:
				servdev = self.inputdev[inputcode]
				if fromgc[2:10] == "DISPINFO":
					statuslist = ("Normal","Idle","Playing","Paused","Fast Forward","Rewind","Play Shuffle","Play Repeat","Play Shuffle Repeat")
					resplist = fromgc.split(",")
					dur = self.secToTime(float(resplist[1][3:]))
					pos = self.secToTime(float(resplist[2][3:]))
					rem = self.secToTime(float(resplist[1][3:])-float(resplist[2][3:]))
					statno = int(resplist[3][-1:])
					stat = statuslist[statno]
					if "Play" in stat:
						self.inputlastst[inputcode] = 0
					else:
						self.inputlastst[inputcode] = round(time.time(),0)
					if servdev.pluginProps["repTrackTime"] == True:
						servdev.updateStateOnServer("tracklen", dur)
						servdev.updateStateOnServer("trackprog", pos)
						servdev.updateStateOnServer("trackrem", rem)
						servdev.updateStateOnServer("status", stat)
				elif fromgc[2:10] == "DISPLINE":
					resplist = fromgc.split(",")
					stateid = "mData" + resplist[0][-1:]
					#linedata = resplist[1][1:-1]
					linedata = fromgc.split("\"")[1].decode('utf_8')
					self.logger.debug(servdev.name + " line " + resplist[0][-1:] + ": " + linedata)
					servdev.updateStateOnServer(stateid, linedata)
			else:
				self.logger.debug(inputcode + " is not yet assigned to a device.")
				
	def commandFromZone(self, zonecode, inputcode, command):
		self.logger.info(command + " button pressed on " + self.zonedev[zonecode].name + " keypad.")
		clientip = self.inputremote[inputcode]['clientip']
		artoken = self.inputremote[inputcode]['artoken']
		dacpid = self.inputremote[inputcode]['dacpid']
		srcremport = self.inputremote[inputcode]['srcremport']
		
		dname = self.inputdev[inputcode].name
		
		if command == "PREV":
			action = "previtem"
		elif command == "NEXT":
			action = "nextitem"
		elif command == "PLAYPAUSE":
			action = "playpause"
		else:
			action = "none"
		
		if srcremport == "":
			self.logger.info("DACP Remote Port not available for " + self.inputdev[inputcode].name)
		elif clientip == "":
			self.logger.info("No client IP available for device streaming to " + self.inputdev[inputcode].name)
			self.logger.info("Try disconnecting and reconnection client from Shairport Sync")
		else:
			self.logger.debug(dname + " IP Address: " + clientip)
			self.logger.debug(dname + " Port: " + srcremport)
			self.logger.debug(dname + " DACP ID: " + dacpid)
			self.logger.debug(dname + " Active Remote Token: " + artoken)
		
			# Get ipv4 address for any ipv6 addresses.
			if ":" in clientip:
				triple = socket.gethostbyaddr(clientip)
				clientip = socket.gethostbyname(triple[0])
		
			url = "http://" + clientip + ":" + srcremport +"/ctrl-int/1/"+action
			self.logger.debug(url)
			
			if action != "none":
				parameters = {'Active-Remote': artoken}
				req = requests.get(url,  headers = parameters)
				req = None
	
	def refreshTrackTimes(self):
		#Loop through all of the defined source devices, and if the user has elected to update the track time info AND
		# the source is actively playing, refresh the track times.
		for inputcode, dev in self.inputdev.items():
			if dev.pluginProps["repTrackTime"] == True:
				status = dev.states["status"]
				if "Play" in status:
					self.commQueue.append(inputcode + "DISPINFO?")
				else:
					pass
			else:
				pass

	def secToTime(self, secondtenths):
		seconds = int(secondtenths/10)
		hrs = divmod(seconds,3600)[0]
		mins = divmod(divmod(seconds,3600)[1],60)[0]
		secs = divmod(seconds,60)[1]
		
		if hrs == 0:
			tfh = ""
		else:
			tfh = str(hrs) + ":"
		
		if hrs == 0:
			tfm = str(mins) + ":"
		else:
			if mins < 10:
				tfm = "0" + str(mins) + ":"
			else:
				tfm = str(mins) + ":"
				
		if secs < 10:
			tfs = "0" + str(secs)
		else:
			tfs = str(secs)
				
		time = tfh + tfm + tfs
		return time
		
	def timeToSec(self, timeString):
		timeparts = timeString.split(":")
		if len(timeparts) == 2:
			seconds = (60 * int(timeparts[0])) + int(timeparts[1])
		elif len(timeparts) == 3:
			seconds = (3600 * int(timeparts[0])) + (60 * int(timeparts[1])) + int(timeparts[2])
		else:
			seconds = 0
		return seconds
			
		
	def actionControlDimmerRelay(self, action, dev):
		if action.deviceAction == indigo.kDeviceAction.TurnOn:
			self.logger.info("Turn On " + dev.name)
			self.logger.debug("Turn On " + dev.pluginProps["zoneselect"])
			self.commQueue.append(dev.pluginProps["zoneselect"] + "ON")
		elif action.deviceAction == indigo.kDeviceAction.TurnOff:
			self.logger.info("Turn Off " + dev.name)
			self.logger.debug("Turn Off " + dev.pluginProps["zoneselect"])
			self.commQueue.append(dev.pluginProps["zoneselect"] + "OFF")
		elif action.deviceAction == indigo.kDeviceAction.Toggle:
			self.logger.info("Toggle " + dev.name)
			self.logger.debug("Toggle " + dev.pluginProps["zoneselect"])
			self.commQueue.append(dev.pluginProps["zoneselect"] + "POWER")
	
	def setZoneVolume(self, pluginAction, dev):
		self.logger.debug("setZoneVolume called with arg: " + str(pluginAction.props.get(u"reqvolume")))
		#if '+' or '-' is specified before the volume setting, do not set it direct, rather increment it by that amount
		#so check for that first.
		if pluginAction.props.get(u"reqvolume")[:1] == "-":
			try:
				newVolume = 79 - (dev.states["Volume"] - int(pluginAction.props.get(u"reqvolume")[1:])) 
				self.logger.info("Decrease volume of " + dev.name + " by " + str(pluginAction.props.get(u"reqvolume")[1:]))
			except ValueError:
				self.logger.info(u"Invalid Volume Level: " + pluginAction.props.get(u"reqvolume"))
				return
		elif pluginAction.props.get(u"reqvolume")[:1] == "+":
			try:
				newVolume = 79 - (dev.states["Volume"] + int(pluginAction.props.get(u"reqvolume")[1:])) 
				self.logger.info("Increase volume of " + dev.name + " by " + str(pluginAction.props.get(u"reqvolume")[1:]))
			except ValueError:
				self.logger.info(u"Invalid Volume Level: " + pluginAction.props.get(u"reqvolume"))
				return
		else:
			self.logger.info("Set volume of " + dev.name + " to " + str(pluginAction.props.get(u"reqvolume")))
			try:
				newVolume = 79 - int(pluginAction.props.get(u"reqvolume"))
			except ValueError:
				self.logger.info(u"Invalid Volume Level: " + pluginAction.props.get(u"reqvolume"))
				return
		
		self.logger.debug("Setting Volume of " + dev.name + " to " + str(pluginAction.props.get(u"reqvolume")))	
		self.commQueue.append(dev.pluginProps["zoneselect"] + "VOL" + str(newVolume))	
		
	def setZoneSource(self, pluginAction, dev):
		## fix Source Name
		inputcode = pluginAction.props.get(u"reqsource").replace("SRC","S")
		if inputcode in self.inputdev:
			sname = self.inputdev[inputcode].name
		else:
			sname = self.inputname[inputcode]
		self.logger.info("Setting Source of " + dev.name + " to " + sname)	
		self.commQueue.append(dev.pluginProps["zoneselect"] + pluginAction.props.get(u"reqsource"))
		
	def setAllOff(self, pluginAction):
		self.commQueue.append("ALLOFF")

	def selectFavorite(self, pluginAction, dev):
		#This code is still active but the XML that calls it is commented out.
		#This code works just fine, but for some reason the GC does not respond to the command
		#as it should. It simply returns '#?'. If I can get it to work, just uncomment the XML and
		#the action becomes available again.
		self.logger.info("Setting Favorite of " + dev.name + " to " + pluginAction.props.get(u"reqfav"))	
		self.commQueue.append(dev.pluginProps["zoneselect"] + "FAV" + pluginAction.props.get(u"reqfav"))

	def sendTextAlert(self, pluginAction):		
		self.logger.debug("sendTextAlert called")
		mtype=pluginAction.props.get(u"msgtype")
		mdwell=pluginAction.props.get(u"msgdwell")
		message=str(pluginAction.props.get(u"messagetext"))
		message=self.substituteVariable(message)
		self.logger.debug(message)
		for x in pluginAction.props.get(u"selectedzones"):
			self.logger.debug("Sending Message Command: " + x + "MSG\"" + message + "\"," + mtype + "," + mdwell)
			self.commQueue.append(x + "MSG\"" + message + "\"," + mtype + "," + mdwell)
			
	def speakAnnouncement(self, pluginAction):		
		self.logger.debug("speakAnnouncement called")
		#First, save current zone parameters
		self.logger.debug("Saving Zone Parameters")
		savedstate = {}
		savedinput = {}
		savedvolume = {}
		for zonecode in pluginAction.props.get(u"selectedzones"):
			dev = self.zonedev[zonecode]
			if dev.states["onOffState"] == True:
				savedstate[zonecode] = "ON"
			else:
				savedstate[zonecode] = "OFF"
			self.logger.debug("Saved State: "+ zonecode + ": " + str(savedstate[zonecode]))
			savedinput[zonecode] = dev.states["Input"]
			self.logger.debug("Saved Input: "+ zonecode + ": " + str(savedinput[zonecode]))
			savedvolume[zonecode] = dev.states["Volume"]
			self.logger.debug("Saved Volume: "+ zonecode + ": " + str(int(savedvolume[zonecode])))
				
		#Now, go through all of the zones and switch them to the right input.
		self.logger.debug("Setting to correct input")
		for zonecode in pluginAction.props.get(u"selectedzones"):
			self.commQueue.append(zonecode + pluginAction.props.get(u"annSource"))
			try:
				newVolume = 79 - int(pluginAction.props.get(u"annvolume"))
			except ValueError:
				self.logger.info(u"Invalid Volume Level: " + pluginAction.props.get(u"annvolume"))
				return
			self.commQueue.append(zonecode + "VOL" + str(newVolume))
			self.commQueue.append(zonecode + "ON")			
		
		#Make Announcement
		time.sleep(2)
		announcetext=pluginAction.props.get("announcement")
		self.logger.debug("Ann Text: " + announcetext)
		announcetext=self.substituteVariable(announcetext)
		self.logger.debug("After Sub, Ann Text: " + announcetext)
		self.logger.debug("Speaking Now...")
		wait = not pluginAction.props.get(u"manualWait")
		waittime = 0
		if pluginAction.props.get(u"manualWait") == True:
			waittime = int(pluginAction.props.get(u"mantime"))
		indigo.server.speak(announcetext, waitUntilDone=wait)
		time.sleep(waittime)
		if pluginAction.props.get(u"airplaydelay") == True:
			time.sleep(2)		
		
		#And finally, reset original parameters
		self.logger.debug("Resetting Orig Zone Parameters")
		for zonecode in pluginAction.props.get(u"selectedzones"):
			self.commQueue.append(zonecode + savedstate[zonecode])
			self.commQueue.append(zonecode + savedinput[zonecode])
			origVolume = 79 - int(savedvolume[zonecode])
			self.commQueue.append(zonecode + "VOL" + str(origVolume))
			